import shutil
from typing import Any, Optional

import pytest

from ..types import ConfigDirSetter, FilePaths
from .integration_data import IntegrationData


class GoWithoutConfigData(IntegrationData):
    def __init__(
        self,
        monkeypatch: pytest.MonkeyPatch,
        set_config_dir: ConfigDirSetter,
        file_paths: FilePaths,
    ) -> None:
        super().__init__(monkeypatch, set_config_dir, file_paths)

    def check_envinronment(self) -> bool:
        return bool(shutil.which("go"))

    @classmethod
    @property
    def input_name(cls) -> str:
        return "GoData"

    def expected_verify_json(self) -> dict[str, Any]:
        return dict(
            {
                "files": {
                    "helloworld.aoj.go": {
                        "additonal_sources": [],
                        "dependencies": [
                            "helloworld.aoj.go",
                            "helloworld/helloworld.go",
                            "helloworld_test.go",
                        ],
                        "document_attributes": {
                            "PROBLEM": "https://onlinejudge.u-aizu.ac.jp/courses/lesson/2/ITP1/1/ITP1_1_A"
                        },
                        "verification": [
                            {
                                "command": {
                                    "command": [
                                        "go",
                                        "run",
                                        f"{self.targets_path}/helloworld.aoj.go",
                                    ],
                                    "env": {"GO111MODULE": "off"},
                                },
                                "name": "go",
                                "problem": "https://onlinejudge.u-aizu.ac.jp/courses/lesson/2/ITP1/1/ITP1_1_A",
                                "type": "problem",
                            }
                        ],
                    },
                    "helloworld/helloworld.go": {
                        "additonal_sources": [],
                        "dependencies": [
                            "helloworld.aoj.go",
                            "helloworld/helloworld.go",
                            "helloworld_test.go",
                        ],
                        "document_attributes": {},
                        "verification": [],
                    },
                    "helloworld_test.go": {
                        "additonal_sources": [],
                        "dependencies": [
                            "helloworld.aoj.go",
                            "helloworld/helloworld.go",
                            "helloworld_test.go",
                        ],
                        "document_attributes": {"UNITTEST": "GOTESTRESULT"},
                        "verification": [{"status": "failure", "type": "const"}],
                    },
                },
            }
        )

    def expected_verify_result(self) -> dict[str, Any]:
        return {
            "files": {
                "helloworld.aoj.go": {
                    "newest": True,
                    "verifications": [
                        {
                            "elapsed": 1415.0,
                            "heaviest": 172.0,
                            "last_execution_time": "2019-03-17T20:56:36.500000-12:00",
                            "slowest": 141.0,
                            "status": "success",
                            "testcases": [
                                {
                                    "elapsed": 4.0,
                                    "memory": 43.12,
                                    "name": "judge_data",
                                    "status": "AC",
                                }
                            ],
                            "verification_name": "go",
                        }
                    ],
                },
                "helloworld_test.go": {
                    "newest": True,
                    "verifications": [
                        {
                            "elapsed": 1938.0,
                            "last_execution_time": "1971-07-29T05:58:46.490000+12:00",
                            "status": "failure",
                        }
                    ],
                },
            },
            "total_seconds": 2547.12,
        }


class GoWithConfigData(GoWithoutConfigData):
    def __init__(
        self,
        monkeypatch: pytest.MonkeyPatch,
        set_config_dir: ConfigDirSetter,
        file_paths: FilePaths,
    ) -> None:
        super().__init__(monkeypatch, set_config_dir, file_paths)

    @classmethod
    @property
    def input_name(cls) -> str:
        return "GoData"

    @property
    def config_path(self) -> Optional[str]:
        return (self.targets_path / "config.toml").as_posix()

    def expected_verify_json(self) -> dict[str, Any]:
        return dict(
            {
                "files": {
                    "helloworld.aoj.go": {
                        "additonal_sources": [],
                        "dependencies": [
                            "helloworld.aoj.go",
                            "helloworld/helloworld.go",
                            "helloworld_test.go",
                        ],
                        "document_attributes": {
                            "PROBLEM": "https://onlinejudge.u-aizu.ac.jp/courses/lesson/2/ITP1/1/ITP1_1_A"
                        },
                        "verification": [
                            {
                                "command": f"{self.config_dir_path/'cache/problems/fbdb181defb159dce09f4dc9338a6728'}/helloworld.aoj.go",
                                "compile": f"env GO111MODULE=off go build -o {self.config_dir_path/'cache/problems/fbdb181defb159dce09f4dc9338a6728'}/helloworld.aoj.go {self.targets_path}/helloworld.aoj.go",
                                "name": "go",
                                "problem": "https://onlinejudge.u-aizu.ac.jp/courses/lesson/2/ITP1/1/ITP1_1_A",
                                "type": "problem",
                            }
                        ],
                    },
                    "helloworld/helloworld.go": {
                        "additonal_sources": [],
                        "dependencies": [
                            "helloworld.aoj.go",
                            "helloworld/helloworld.go",
                            "helloworld_test.go",
                        ],
                        "document_attributes": {},
                        "verification": [],
                    },
                    "helloworld_test.go": {
                        "additonal_sources": [],
                        "dependencies": [
                            "helloworld.aoj.go",
                            "helloworld/helloworld.go",
                            "helloworld_test.go",
                        ],
                        "document_attributes": {"UNITTEST": "GOTESTRESULT"},
                        "verification": [{"status": "failure", "type": "const"}],
                    },
                },
            }
        )



from .file import (
    AddtionalSource,
    DocumentOutputMode,
    VerificationFile,
    VerificationInput,
)
from .path import ForcePosixPath, RelativeDirectoryPath, SortedPathList, SortedPathSet
from .result import (
    FileResult,
    JudgeStatus,
    TestcaseResult,
    VerificationResult,
    VerifyCommandResult,
)
from .result_status import ResultStatus
from .shell import ShellCommand, ShellCommandLike
from .verification import (
    BaseVerification,
    CommandVerification,
    ConstVerification,
    ProblemVerification,
    Verification,
    VerificationParams,
)

__all__ = [
    "ForcePosixPath",
    "SortedPathSet",
    "SortedPathList",
    "RelativeDirectoryPath",
    "VerificationFile",
    "VerificationInput",
    "AddtionalSource",
    "FileResult",
    "ResultStatus",
    "VerifyCommandResult",
    "TestcaseResult",
    "JudgeStatus",
    "BaseVerification",
    "ShellCommand",
    "ShellCommandLike",
    "CommandVerification",
    "ConstVerification",
    "ProblemVerification",
    "Verification",
    "VerificationResult",
    "VerificationParams",
    "DocumentOutputMode",
]



import pathlib
from datetime import datetime, timedelta, timezone
from typing import Any

import pytest

from competitive_verifier.models import (
    FileResult,
    ResultStatus,
    VerificationResult,
    VerifyCommandResult,
)

test_parse_FileResult_params: list[
    tuple[FileResult, dict[str, Any], dict[str, Any], str]
] = [
    (
        FileResult(
            verifications=[
                VerificationResult(
                    elapsed=1.5,
                    status=ResultStatus.SUCCESS,
                    last_execution_time=datetime(2016, 12, 24, 15, 16, 34),
                )
            ],
        ),
        {
            "verifications": [
                {
                    "status": "success",
                    "elapsed": 1.5,
                    "last_execution_time": "2016-12-24 15:16:34",
                }
            ],
            "newest": True,
        },
        {
            "verifications": [
                {
                    "status": "success",
                    "elapsed": 1.5,
                    "last_execution_time": datetime(2016, 12, 24, 15, 16, 34),
                }
            ],
            "newest": True,
        },
        '{"verifications":[{"status":"success","elapsed":1.5,"last_execution_time":"2016-12-24T15:16:34"}],"newest":true}',
    ),
    (
        FileResult(
            newest=False,
            verifications=[
                VerificationResult(
                    elapsed=1.5,
                    status=ResultStatus.SUCCESS,
                    last_execution_time=datetime(
                        2016, 12, 24, 15, 16, 34, tzinfo=timezone.utc
                    ),
                )
            ],
        ),
        {
            "newest": False,
            "verifications": [
                {
                    "status": "Success",
                    "elapsed": 1.5,
                    "last_execution_time": "2016-12-24 15:16:34Z",
                }
            ],
        },
        {
            "verifications": [
                {
                    "status": "success",
                    "elapsed": 1.5,
                    "last_execution_time": datetime(
                        2016, 12, 24, 15, 16, 34, tzinfo=timezone.utc
                    ),
                }
            ],
            "newest": False,
        },
        '{"verifications":[{"status":"success","elapsed":1.5,"last_execution_time":"2016-12-24T15:16:34Z"}],"newest":false}',
    ),
    (
        FileResult(
            newest=True,
            verifications=[
                VerificationResult(
                    elapsed=1.5,
                    status=ResultStatus.SUCCESS,
                    last_execution_time=datetime(
                        2016, 12, 24, 15, 16, 34, tzinfo=timezone(timedelta(hours=9))
                    ),
                )
            ],
        ),
        {
            "verifications": [
                {
                    "elapsed": 1.5,
                    "status": "SUCCESS",
                    "last_execution_time": "2016-12-24 15:16:34+09:00",
                }
            ]
        },
        {
            "verifications": [
                {
                    "elapsed": 1.5,
                    "status": "success",
                    "last_execution_time": datetime(
                        2016, 12, 24, 15, 16, 34, tzinfo=timezone(timedelta(hours=9))
                    ),
                }
            ],
            "newest": True,
        },
        '{"verifications":[{"status":"success","elapsed":1.5,"last_execution_time":"2016-12-24T15:16:34+09:00"}],"newest":true}',
    ),
]


@pytest.mark.parametrize(
    "obj, raw_dict, output_dict, output_json",
    test_parse_FileResult_params,
)
def test_parse_FileResult(
    obj: FileResult,
    raw_dict: dict[str, Any],
    output_dict: dict[str, Any],
    output_json: str,
):
    assert obj == FileResult.model_validate(raw_dict)
    assert obj.model_dump(exclude_none=True) == output_dict
    assert obj.model_dump_json(exclude_none=True) == output_json


test_file_result_need_verification_params: list[tuple[FileResult, datetime, bool]] = [
    (FileResult(verifications=[]), datetime(2016, 12, 24, 19, 0, 0), True),
    (
        FileResult(
            verifications=[
                VerificationResult(
                    elapsed=1.5,
                    status=ResultStatus.SUCCESS,
                    last_execution_time=datetime(2019, 12, 24, 19, 0, 0),
                )
            ]
        ),
        datetime(2016, 12, 24, 19, 0, 0),
        False,
    ),
    (
        FileResult(
            verifications=[
                VerificationResult(
                    elapsed=1.5,
                    status=ResultStatus.FAILURE,
                    last_execution_time=datetime(2019, 12, 24, 19, 0, 0),
                )
            ]
        ),
        datetime(2016, 12, 24, 19, 0, 0),
        True,
    ),
    (
        FileResult(
            verifications=[
                VerificationResult(
                    elapsed=1.5,
                    status=ResultStatus.SKIPPED,
                    last_execution_time=datetime(2019, 12, 24, 19, 0, 0),
                )
            ]
        ),
        datetime(2016, 12, 24, 19, 0, 0),
        True,
    ),
    (
        FileResult(
            verifications=[
                VerificationResult(
                    elapsed=1.5,
                    status=ResultStatus.SUCCESS,
                    last_execution_time=datetime(2015, 12, 24, 19, 0, 0),
                )
            ]
        ),
        datetime(2016, 12, 24, 19, 0, 0),
        True,
    ),
    (
        FileResult(
            verifications=[
                VerificationResult(
                    elapsed=1.5,
                    status=ResultStatus.SUCCESS,
                    last_execution_time=datetime(2018, 12, 24, 19, 0, 0),
                ),
                VerificationResult(
                    elapsed=1.5,
                    status=ResultStatus.SUCCESS,
                    last_execution_time=datetime(2018, 12, 24, 19, 0, 0),
                ),
                VerificationResult(
                    elapsed=1.5,
                    status=ResultStatus.SUCCESS,
                    last_execution_time=datetime(2015, 12, 24, 19, 0, 0),
                ),
            ]
        ),
        datetime(2016, 12, 24, 19, 0, 0),
        True,
    ),
    (
        FileResult(
            verifications=[
                VerificationResult(
                    elapsed=1.5,
                    status=ResultStatus.SUCCESS,
                    last_execution_time=datetime(2018, 12, 24, 19, 0, 0),
                ),
                VerificationResult(
                    elapsed=1.5,
                    status=ResultStatus.SUCCESS,
                    last_execution_time=datetime(2018, 12, 24, 19, 0, 0),
                ),
                VerificationResult(
                    elapsed=1.5,
                    status=ResultStatus.FAILURE,
                    last_execution_time=datetime(2018, 12, 24, 19, 0, 0),
                ),
            ]
        ),
        datetime(2016, 12, 24, 19, 0, 0),
        True,
    ),
]


@pytest.mark.parametrize(
    "obj, dt, expected",
    test_file_result_need_verification_params,
)
def test_file_result_need_verification(
    obj: FileResult,
    dt: datetime,
    expected: bool,
):
    assert obj.need_verification(dt) == expected


test_is_success_params = [
    (
        FileResult(
            verifications=[
                VerificationResult(elapsed=1, status=ResultStatus.SUCCESS),
                VerificationResult(elapsed=1, status=ResultStatus.SUCCESS),
            ]
        ),
        (True, True),
    ),
    (
        FileResult(
            verifications=[
                VerificationResult(elapsed=1, status=ResultStatus.SUCCESS),
                VerificationResult(elapsed=1, status=ResultStatus.FAILURE),
            ]
        ),
        (False, False),
    ),
    (
        FileResult(
            verifications=[
                VerificationResult(elapsed=1, status=ResultStatus.SUCCESS),
                VerificationResult(elapsed=1, status=ResultStatus.SKIPPED),
            ]
        ),
        (False, True),
    ),
]


@pytest.mark.parametrize(
    "obj, expected",
    test_is_success_params,
)
def test_is_success(
    obj: FileResult,
    expected: tuple[bool, bool],
):
    assert obj.is_success(allow_skip=False) == expected[0]
    assert obj.is_success(allow_skip=True) == expected[1]


def test_verify_command_result_json():
    obj = VerifyCommandResult(
        total_seconds=3.75,
        files={
            pathlib.Path("foo/bar.py"): FileResult(
                verifications=[
                    VerificationResult(elapsed=1, status=ResultStatus.SUCCESS),
                    VerificationResult(elapsed=1, status=ResultStatus.SKIPPED),
                ]
            ),
            pathlib.Path("foo/baz.py"): FileResult(
                verifications=[
                    VerificationResult(elapsed=1, status=ResultStatus.SUCCESS),
                ]
            ),
        },
    )
    assert (
        VerifyCommandResult.model_validate_json(obj.model_dump_json(exclude_none=True))
        == obj
    )


test_merge_params = [
    (
        VerifyCommandResult(
            total_seconds=4.25,
            files={
                pathlib.Path("foo/bar.py"): FileResult(
                    verifications=[
                        VerificationResult(
                            elapsed=1,
                            status=ResultStatus.SUCCESS,
                            last_execution_time=datetime.fromtimestamp(1675125600),
                        ),
                        VerificationResult(
                            elapsed=2,
                            status=ResultStatus.SKIPPED,
                            last_execution_time=datetime.fromtimestamp(1675125601),
                        ),
                    ]
                ),
                pathlib.Path("foo/baz.py"): FileResult(
                    verifications=[
                        VerificationResult(
                            elapsed=3,
                            status=ResultStatus.SUCCESS,
                            last_execution_time=datetime.fromtimestamp(1675125602),
                        ),
                    ]
                ),
            },
        ),
        VerifyCommandResult(
            total_seconds=3,
            files={
                pathlib.Path("foo/bar.py"): FileResult(
                    verifications=[
                        VerificationResult(
                            elapsed=4,
                            status=ResultStatus.SUCCESS,
                            last_execution_time=datetime.fromtimestamp(1675125603),
                        ),
                        VerificationResult(
                            elapsed=5,
                            status=ResultStatus.SKIPPED,
                            last_execution_time=datetime.fromtimestamp(1675125604),
                        ),
                    ]
                ),
                pathlib.Path("foo/baz2.py"): FileResult(
                    verifications=[
                        VerificationResult(
                            elapsed=6,
                            status=ResultStatus.SUCCESS,
                            last_execution_time=datetime.fromtimestamp(1675125605),
                        ),
                    ]
                ),
            },
        ),
        VerifyCommandResult(
            total_seconds=7.25,
            files={
                pathlib.Path("foo/bar.py"): FileResult(
                    verifications=[
                        VerificationResult(
                            elapsed=4,
                            status=ResultStatus.SUCCESS,
                            last_execution_time=datetime.fromtimestamp(1675125603),
                        ),
                        VerificationResult(
                            elapsed=5,
                            status=ResultStatus.SKIPPED,
                            last_execution_time=datetime.fromtimestamp(1675125604),
                        ),
                    ]
                ),
                pathlib.Path("foo/baz.py"): FileResult(
                    verifications=[
                        VerificationResult(
                            elapsed=3,
                            status=ResultStatus.SUCCESS,
                            last_execution_time=datetime.fromtimestamp(1675125602),
                        ),
                    ]
                ),
                pathlib.Path("foo/baz2.py"): FileResult(
                    verifications=[
                        VerificationResult(
                            elapsed=6,
                            status=ResultStatus.SUCCESS,
                            last_execution_time=datetime.fromtimestamp(1675125605),
                        ),
                    ]
                ),
            },
        ),
    ),
    (
        VerifyCommandResult(
            total_seconds=4.25,
            files={
                pathlib.Path("foo/bar.py"): FileResult(
                    verifications=[
                        VerificationResult(
                            elapsed=1,
                            status=ResultStatus.SUCCESS,
                            last_execution_time=datetime.fromtimestamp(1675125600),
                        ),
                        VerificationResult(
                            elapsed=2,
                            status=ResultStatus.SKIPPED,
                            last_execution_time=datetime.fromtimestamp(1675125601),
                        ),
                    ]
                ),
                pathlib.Path("foo/baz.py"): FileResult(
                    verifications=[
                        VerificationResult(
                            elapsed=3,
                            status=ResultStatus.SUCCESS,
                            last_execution_time=datetime.fromtimestamp(1675125602),
                        ),
                    ],
                    newest=True,
                ),
            },
        ),
        VerifyCommandResult(
            total_seconds=3,
            files={
                pathlib.Path("foo/bar.py"): FileResult(
                    verifications=[
                        VerificationResult(
                            elapsed=4,
                            status=ResultStatus.SUCCESS,
                            last_execution_time=datetime.fromtimestamp(1675125603),
                        ),
                        VerificationResult(
                            elapsed=5,
                            status=ResultStatus.SKIPPED,
                            last_execution_time=datetime.fromtimestamp(1675125604),
                        ),
                    ]
                ),
                pathlib.Path("foo/baz.py"): FileResult(
                    verifications=[
                        VerificationResult(
                            elapsed=6,
                            status=ResultStatus.SUCCESS,
                            last_execution_time=datetime.fromtimestamp(1675125605),
                        ),
                    ],
                    newest=False,
                ),
            },
        ),
        VerifyCommandResult(
            total_seconds=7.25,
            files={
                pathlib.Path("foo/bar.py"): FileResult(
                    verifications=[
                        VerificationResult(
                            elapsed=4,
                            status=ResultStatus.SUCCESS,
                            last_execution_time=datetime.fromtimestamp(1675125603),
                        ),
                        VerificationResult(
                            elapsed=5,
                            status=ResultStatus.SKIPPED,
                            last_execution_time=datetime.fromtimestamp(1675125604),
                        ),
                    ]
                ),
                pathlib.Path("foo/baz.py"): FileResult(
                    verifications=[
                        VerificationResult(
                            elapsed=3,
                            status=ResultStatus.SUCCESS,
                            last_execution_time=datetime.fromtimestamp(1675125602),
                        ),
                    ]
                ),
            },
        ),
    ),
    (
        VerifyCommandResult(
            total_seconds=4.25,
            files={
                pathlib.Path("foo/bar.py"): FileResult(
                    verifications=[
                        VerificationResult(
                            elapsed=1,
                            status=ResultStatus.SUCCESS,
                            last_execution_time=datetime.fromtimestamp(1675125600),
                        ),
                        VerificationResult(
                            elapsed=2,
                            status=ResultStatus.SKIPPED,
                            last_execution_time=datetime.fromtimestamp(1675125601),
                        ),
                    ]
                ),
                pathlib.Path("foo/baz.py"): FileResult(
                    verifications=[
                        VerificationResult(
                            elapsed=3,
                            status=ResultStatus.SUCCESS,
                            last_execution_time=datetime.fromtimestamp(1675125602),
                        ),
                    ],
                    newest=False,
                ),
            },
        ),
        VerifyCommandResult(
            total_seconds=3,
            files={
                pathlib.Path("foo/bar.py"): FileResult(
                    verifications=[
                        VerificationResult(
                            elapsed=4,
                            status=ResultStatus.SUCCESS,
                            last_execution_time=datetime.fromtimestamp(1675125603),
                        ),
                        VerificationResult(
                            elapsed=5,
                            status=ResultStatus.SKIPPED,
                            last_execution_time=datetime.fromtimestamp(1675125604),
                        ),
                    ]
                ),
                pathlib.Path("foo/baz.py"): FileResult(
                    verifications=[
                        VerificationResult(
                            elapsed=6,
                            status=ResultStatus.SUCCESS,
                            last_execution_time=datetime.fromtimestamp(1675125605),
                        ),
                    ],
                    newest=True,
                ),
            },
        ),
        VerifyCommandResult(
            total_seconds=7.25,
            files={
                pathlib.Path("foo/bar.py"): FileResult(
                    verifications=[
                        VerificationResult(
                            elapsed=4,
                            status=ResultStatus.SUCCESS,
                            last_execution_time=datetime.fromtimestamp(1675125603),
                        ),
                        VerificationResult(
                            elapsed=5,
                            status=ResultStatus.SKIPPED,
                            last_execution_time=datetime.fromtimestamp(1675125604),
                        ),
                    ]
                ),
                pathlib.Path("foo/baz.py"): FileResult(
                    verifications=[
                        VerificationResult(
                            elapsed=6,
                            status=ResultStatus.SUCCESS,
                            last_execution_time=datetime.fromtimestamp(1675125605),
                        ),
                    ]
                ),
            },
        ),
    ),
    (
        VerifyCommandResult(
            total_seconds=4.25,
            files={
                pathlib.Path("foo/bar.py"): FileResult(
                    verifications=[
                        VerificationResult(
                            elapsed=1,
                            status=ResultStatus.SUCCESS,
                            last_execution_time=datetime.fromtimestamp(1675125600),
                        ),
                        VerificationResult(
                            elapsed=2,
                            status=ResultStatus.SKIPPED,
                            last_execution_time=datetime.fromtimestamp(1675125601),
                        ),
                    ]
                ),
                pathlib.Path("foo/baz.py"): FileResult(
                    verifications=[
                        VerificationResult(
                            elapsed=3,
                            status=ResultStatus.SUCCESS,
                            last_execution_time=datetime.fromtimestamp(1675125602),
                        ),
                    ],
                    newest=True,
                ),
            },
        ),
        VerifyCommandResult(
            total_seconds=3,
            files={
                pathlib.Path("foo/bar.py"): FileResult(
                    verifications=[
                        VerificationResult(
                            elapsed=4,
                            status=ResultStatus.SUCCESS,
                            last_execution_time=datetime.fromtimestamp(1675125603),
                        ),
                        VerificationResult(
                            elapsed=5,
                            status=ResultStatus.SKIPPED,
                            last_execution_time=datetime.fromtimestamp(1675125604),
                        ),
                    ]
                ),
                pathlib.Path("foo/baz.py"): FileResult(
                    verifications=[
                        VerificationResult(
                            elapsed=6,
                            status=ResultStatus.SUCCESS,
                            last_execution_time=datetime.fromtimestamp(1675125605),
                        ),
                    ],
                    newest=True,
                ),
            },
        ),
        VerifyCommandResult(
            total_seconds=7.25,
            files={
                pathlib.Path("foo/bar.py"): FileResult(
                    verifications=[
                        VerificationResult(
                            elapsed=4,
                            status=ResultStatus.SUCCESS,
                            last_execution_time=datetime.fromtimestamp(1675125603),
                        ),
                        VerificationResult(
                            elapsed=5,
                            status=ResultStatus.SKIPPED,
                            last_execution_time=datetime.fromtimestamp(1675125604),
                        ),
                    ]
                ),
                pathlib.Path("foo/baz.py"): FileResult(
                    verifications=[
                        VerificationResult(
                            elapsed=6,
                            status=ResultStatus.SUCCESS,
                            last_execution_time=datetime.fromtimestamp(1675125605),
                        ),
                    ]
                ),
            },
        ),
    ),
]


@pytest.mark.parametrize(
    "obj1, obj2, expected",
    test_merge_params,
)
def test_merge(
    obj1: VerifyCommandResult,
    obj2: VerifyCommandResult,
    expected: VerifyCommandResult,
):
    assert obj1.merge(obj2) == expected



import argparse
import logging
import pathlib
import sys
from functools import reduce
from logging import getLogger
from typing import Iterable, Optional

from competitive_verifier import github, summary
from competitive_verifier.arg import (
    add_result_json_argument,
    add_verbose_argument,
    add_write_summary_argument,
)
from competitive_verifier.log import configure_stderr_logging
from competitive_verifier.models import VerifyCommandResult

logger = getLogger(__name__)


def merge(results: Iterable[VerifyCommandResult]) -> VerifyCommandResult:
    return reduce(lambda a, b: a.merge(b), results)


def run_impl(
    *result_json: pathlib.Path,
    write_summary: bool = False,
) -> VerifyCommandResult:
    configure_stderr_logging()
    result = merge(map(VerifyCommandResult.parse_file_relative, result_json))
    if write_summary:
        gh_summary_path = github.env.get_step_summary_path()
        if gh_summary_path and gh_summary_path.parent.exists():
            with open(gh_summary_path, "w", encoding="utf-8") as fp:
                summary.write_summary(fp, result)
        else:
            logger.warning("write_summary=True but not found $GITHUB_STEP_SUMMARY")
    return result


def run(args: argparse.Namespace) -> bool:
    default_level = logging.INFO
    if args.verbose:
        default_level = logging.DEBUG
    configure_stderr_logging(default_level)

    merged = run_impl(*args.result_json, write_summary=args.write_summary)
    print(merged.model_dump_json(exclude_none=True))
    return True


def argument(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    add_verbose_argument(parser)
    add_result_json_argument(parser)
    add_write_summary_argument(parser)
    return parser


def main(args: Optional[list[str]] = None) -> None:
    try:
        parsed = argument(argparse.ArgumentParser()).parse_args(args)
        if not run(parsed):
            sys.exit(1)
    except Exception as e:
        sys.stderr.write(str(e))
        sys.exit(2)


if __name__ == "__main__":
    main()



# Python Version: 3.x
import pathlib
from logging import getLogger
from tempfile import TemporaryDirectory
from typing import Optional, Sequence, Union

from pydantic import BaseModel

import competitive_verifier.oj.verify.languages.special_comments as special_comments
import competitive_verifier.oj.verify.utils as utils
from competitive_verifier.models import ShellCommand, ShellCommandLike
from competitive_verifier.oj.verify.models import (
    Language,
    LanguageEnvironment,
    OjVerifyUserDefinedConfig,
)

logger = getLogger(__name__)

StrPath = Union[pathlib.Path, str]


class PathContainer(BaseModel):
    path: StrPath
    basedir: StrPath
    tempdir: StrPath

    def _format(self, input: str):
        return input.format(
            path=str(self.path),
            dir=str(pathlib.Path(self.path).parent),
            basedir=str(self.basedir),
            tempdir=str(self.tempdir),
        )

    def format_command(self, command: ShellCommandLike) -> ShellCommandLike:
        if isinstance(command, str):
            return self._format(command)

        if not isinstance(command, ShellCommand):
            return list(map(self._format, command))

        if isinstance(command.command, str):
            cmd = self._format(command.command)
        else:
            cmd = list(map(self._format, command.command))

        env = (
            {self._format(k): self._format(v) for k, v in command.env.items()}
            if command.env
            else None
        )

        cwd = pathlib.Path(self._format(str(command.cwd))) if command.cwd else None

        return ShellCommand(
            command=cmd,
            env=env,
            cwd=cwd,
        )

    def parse_command(self, command: ShellCommandLike) -> ShellCommand:
        return ShellCommand.parse_command_like(self.format_command(command))


class UserDefinedLanguageEnvironment(LanguageEnvironment):
    config: OjVerifyUserDefinedConfig
    _name: str

    def __init__(self, *, name: str, config: OjVerifyUserDefinedConfig):
        self.config = config
        self._name = name

    @property
    def name(self) -> str:
        return self._name

    def get_compile_command(
        self, path: pathlib.Path, *, basedir: pathlib.Path, tempdir: pathlib.Path
    ) -> Optional[ShellCommandLike]:
        if self.config.compile:
            return PathContainer(
                path=path, basedir=basedir, tempdir=tempdir
            ).format_command(self.config.compile)
        return None

    def get_execute_command(
        self, path: pathlib.Path, *, basedir: pathlib.Path, tempdir: pathlib.Path
    ) -> ShellCommandLike:
        return PathContainer(
            path=path, basedir=basedir, tempdir=tempdir
        ).format_command(self.config.execute)


class UserDefinedLanguage(Language):
    extension: str
    config: OjVerifyUserDefinedConfig

    def __init__(self, *, extension: str, config: OjVerifyUserDefinedConfig):
        self.extension = extension
        self.config = config

    def list_attributes(
        self, path: pathlib.Path, *, basedir: pathlib.Path
    ) -> dict[str, str]:
        if self.config.list_attributes is None:
            return dict(special_comments.list_special_comments(path))

        with TemporaryDirectory() as tempdir:
            text = (
                PathContainer(path=path, basedir=basedir, tempdir=tempdir)
                .parse_command(self.config.list_attributes)
                .exec_command(text=False, capture_output=True)
                .stdout
            )
        attributes: dict[str, str] = {}
        for line in text.splitlines():
            key, _, value = line.decode().partition(" ")
            attributes[key] = value
        return attributes

    def list_dependencies(
        self, path: pathlib.Path, *, basedir: pathlib.Path
    ) -> list[pathlib.Path]:
        if self.config.list_dependencies is None:
            logger.warning(
                "The functionality to list dependencies of .%s file is not implemented yet.",
                self.extension,
            )
            return list(
                utils.glob_with_predicate(
                    lambda path: path.suffix == "." + self.extension
                )
            )

        with TemporaryDirectory() as tempdir:
            text = (
                PathContainer(path=path, basedir=basedir, tempdir=tempdir)
                .parse_command(self.config.list_dependencies)
                .exec_command(text=False, capture_output=True)
                .stdout
            )
        dependencies = [path]
        for line in text.splitlines():
            dependencies.append(pathlib.Path(line.decode()))
        return dependencies

    def bundle(self, path: pathlib.Path, *, basedir: pathlib.Path) -> Optional[bytes]:
        if self.config.bundle is None:
            return None
        with TemporaryDirectory() as tempdir:
            return (
                PathContainer(path=path, basedir=basedir, tempdir=tempdir)
                .parse_command(self.config.bundle)
                .exec_command(text=False, capture_output=True)
                .stdout
            )

    def list_environments(
        self, path: pathlib.Path, *, basedir: pathlib.Path
    ) -> Sequence[LanguageEnvironment]:
        return [UserDefinedLanguageEnvironment(name=self.extension, config=self.config)]



import os
import pathlib
import shutil
import textwrap
from contextlib import nullcontext
from logging import getLogger
from typing import Iterator, Optional

import onlinejudge.dispatch as dispatch
import onlinejudge_command.download_history
import onlinejudge_command.format_utils as format_utils
import onlinejudge_command.main
import onlinejudge_command.pretty_printers as pretty_printers
import onlinejudge_command.utils as utils
import requests.exceptions
from onlinejudge.service.atcoder import AtCoderProblem
from onlinejudge.service.yukicoder import YukicoderProblem
from onlinejudge.type import NotLoggedInError, SampleParseError, TestCase

from competitive_verifier import log

from .func import (
    get_cache_directory,
    get_checker_path,
    get_directory,
    is_atcoder,
    is_yukicoder,
)

logger = getLogger(__name__)


# flake8: noqa: C901
def run(
    *,
    url: str,
    directory: pathlib.Path,
    cookie: pathlib.Path,
    format: Optional[str] = None,
    system: bool = True,
    silent: bool = True,
    dry_run: bool = False,
) -> bool:
    # prepare values
    problem = dispatch.problem_from_url(url)
    if problem is None:
        if dispatch.contest_from_url(url) is not None:
            logger.warning(
                "You specified a URL for a contest instead of a problem. If you want to download for all problems of a contest at once, please try to use `oj-prepare` command of https://github.com/online-judge-tools/template-generator"
            )
        logger.error('The URL "%s" is not supported', url)
        return False

    is_default_format = format is None
    if format is None:
        format = "%b.%e"

    # get samples from the server
    with utils.new_session_with_our_user_agent(path=cookie) as sess:
        if isinstance(problem, AtCoderProblem) and system:
            DROPBOX_TOKEN = os.environ.get("DROPBOX_TOKEN")
            if not DROPBOX_TOKEN:
                logger.info(
                    utils.HINT
                    + "You need to give the access token. Please do the following:\n%s",
                    textwrap.dedent(
                        """
                        1. Open the following URL in your browser:
                            https://www.dropbox.com/oauth2/authorize?client_id=153gig8dqgk3ujg&response_type=code
                        2. Authorize the app and take the access code.
                        3. Run the following command with replacing the "${YOUR_ACCESS_CODE}":
                            $ curl https://api.dropbox.com/oauth2/token --user 153gig8dqgk3ujg:5l7o7lh73o8i9ux --data grant_type=authorization_code --data code=${YOUR_ACCESS_CODE}
                        4. Get the access token from the JSON. It is in the "access_token" field.
                        5. Use the access token. For example:
                            $ oj download """
                        + problem.get_url()
                        + """ --system --dropbox-token=${YOUR_ACCESS_TOKEN}

                    (Please take care that the access code and the access token are CONFIDENTIAL information. DON'T SHARE with other people!)
                """
                    ),
                )
                raise SampleParseError("--dropbox-token is not given")
            sess.headers["Authorization"] = "Bearer {}".format(DROPBOX_TOKEN)
        if isinstance(problem, YukicoderProblem):
            YUKICODER_TOKEN = os.environ.get("YUKICODER_TOKEN")
            if YUKICODER_TOKEN:
                sess.headers["Authorization"] = "Bearer {}".format(YUKICODER_TOKEN)
        try:
            if system:
                samples = problem.download_system_cases(session=sess)
            else:
                samples = problem.download_sample_cases(session=sess)
        except requests.exceptions.RequestException as e:
            logger.error("%s", e)
            logger.error(
                utils.HINT
                + "You may need to login to use `$ oj download ...` during contest. Please run: $ oj login %s",
                problem.get_service().get_url(),
            )
            return False
        except SampleParseError as e:
            logger.error("%s", e)
            return False

    if not samples:
        logger.error("Sample not found")
        return False

    # append the history for submit subcommand
    if not dry_run and is_default_format:
        history = onlinejudge_command.download_history.DownloadHistory()
        if not list(directory.glob("*")):
            # reset the history to help users who use only one directory for many problems
            history.remove(directory=pathlib.Path.cwd())
        history.add(problem, directory=pathlib.Path.cwd())

    # prepare files to write
    def iterate_files_to_write(
        sample: TestCase, *, i: int
    ) -> Iterator[tuple[str, pathlib.Path, bytes]]:
        for ext in ["in", "out"]:
            data = getattr(sample, ext + "put_data")
            if data is None:
                continue
            name = sample.name
            table = {}
            table["i"] = str(i + 1)
            table["e"] = ext
            table["n"] = name
            table["b"] = os.path.basename(name)
            table["d"] = os.path.dirname(name)
            path: pathlib.Path = directory / format_utils.percentformat(format, table)
            yield ext, path, data

    for i, sample in enumerate(samples):
        for _, path, _ in iterate_files_to_write(sample, i=i):
            if path.exists():
                logger.error(
                    "Failed to download since file already exists: %s", str(path)
                )
                logger.info(
                    utils.HINT
                    + "We recommend adding your own test cases to test/ directory, and using one directory per one problem. Please see also https://github.com/online-judge-tools/oj/blob/master/docs/getting-started.md#random-testing. If you wanted to keep using one directory per one contest, you can run like `$ rm -rf test/ && oj d https://...`."
                )
                return False

    # write samples to files
    for i, sample in enumerate(samples):
        logger.info("")
        logger.info("sample %d", i)
        for ext, path, data in iterate_files_to_write(sample, i=i):
            content = ""
            if not silent:
                content = "\n" + pretty_printers.make_pretty_large_file_content(
                    data, limit=40, head=20, tail=10
                )
            logger.info("%sput: %s%s", ext, sample.name, content)
            if not dry_run:
                path.parent.mkdir(parents=True, exist_ok=True)
                with path.open("wb") as fh:
                    fh.write(data)
                logger.info(utils.SUCCESS + "saved to: %s", path)

    return True


def run_wrapper(url: str, *, group_log: bool = False) -> bool:
    directory = get_directory(url)
    test_directory = directory / "test"

    logger.info("download[Start]: %s into %s", url, test_directory)
    if not (test_directory).exists() or list((test_directory).iterdir()) == []:
        logger.info("download[Run]: %s", url)

        if group_log:
            cm = log.group(f"download[Run]: {url}")
        else:
            cm = nullcontext()
        with cm:
            directory.mkdir(parents=True, exist_ok=True)
            # time.sleep(2)

            try:
                run(
                    url=url,
                    directory=test_directory,
                    cookie=get_cache_directory() / "cookie.txt",
                )
                checker_path = get_checker_path(url)
                if checker_path and checker_path.exists():
                    try:
                        shutil.copy2(checker_path, directory)
                    except Exception as e:
                        logger.exception("Failed to copy checker %s", e)
                        shutil.rmtree(directory)
                        return False
            except Exception as e:
                if isinstance(e, NotLoggedInError) and is_yukicoder(url):
                    logger.error("Required: $YUKICODER_TOKEN environment variable")
                elif isinstance(e, NotLoggedInError) and is_atcoder(url):
                    logger.error("Required: $DROPBOX_TOKEN environment variable")
                else:
                    logger.exception("Failed to download: %s", e)
                return False
    else:
        logger.info("download:already exists: %s", url)
    return True



import pathlib
from datetime import datetime, timedelta
from io import StringIO

import pytest

import competitive_verifier.summary as summary
from competitive_verifier.models import FileResult, JudgeStatus, ResultStatus
from competitive_verifier.models import TestcaseResult as CaseResult
from competitive_verifier.models import VerificationResult, VerifyCommandResult

test_to_human_str_seconds_params: list[tuple[timedelta, str]] = [
    (timedelta(hours=2, minutes=8, seconds=6, milliseconds=13), "2h 8m"),
    (timedelta(minutes=8, seconds=6, milliseconds=13), "8m 6s"),
    (timedelta(seconds=16, milliseconds=13), "16s"),
    (timedelta(seconds=6, milliseconds=3), "6.0s"),
    (timedelta(seconds=6, milliseconds=13), "6.0s"),
    (timedelta(seconds=6, milliseconds=213), "6.2s"),
    (timedelta(milliseconds=1), "1ms"),
    (timedelta(milliseconds=12), "12ms"),
    (timedelta(milliseconds=123), "123ms"),
]


@pytest.mark.parametrize(
    "td, expected",
    test_to_human_str_seconds_params,
    ids=str,
)
def test_to_human_str_seconds(td: timedelta, expected: str):
    assert summary.to_human_str_seconds(td.total_seconds()) == expected


test_to_human_str_mega_bytes_params: list[tuple[float, str]] = [
    (0.000123456789, "0MB"),
    (0.00123456, "0.00123MB"),
    (0.01234567, "0.0123MB"),
    (0.12345678, "0.123MB"),
    (1.23456789, "1.23MB"),
    (12.3456789, "12.3MB"),
    (123.456789, "123MB"),
    (1234.56789, "1234MB"),
    (12345.6789, "12345MB"),
]


@pytest.mark.parametrize(
    "megabytes, expected",
    test_to_human_str_mega_bytes_params,
)
def test_to_human_str_mega_bytes(megabytes: float, expected: str):
    assert summary.to_human_str_mega_bytes(megabytes) == expected


def test_summary():
    with StringIO() as fp:
        summary.write_summary(
            fp=fp,
            result=VerifyCommandResult(
                total_seconds=4.25,
                files={
                    pathlib.Path("foo/bar.py"): FileResult(
                        verifications=[
                            VerificationResult(
                                verification_name="ver",
                                heaviest=31.432342,
                                slowest=0.420321,
                                elapsed=1,
                                status=ResultStatus.FAILURE,
                                last_execution_time=datetime.fromtimestamp(1675125600),
                                testcases=[
                                    CaseResult(
                                        name="case01",
                                        status=JudgeStatus.TLE,
                                        elapsed=0.2,
                                        memory=30,
                                    ),
                                    CaseResult(
                                        name="case02",
                                        status=JudgeStatus.WA,
                                        elapsed=0.4,
                                        memory=32,
                                    ),
                                ],
                            ),
                            VerificationResult(
                                elapsed=2,
                                status=ResultStatus.FAILURE,
                                last_execution_time=datetime.fromtimestamp(1675125601),
                            ),
                        ]
                    ),
                    pathlib.Path("foo/baz.py"): FileResult(
                        verifications=[
                            VerificationResult(
                                verification_name="fi",
                                heaviest=31.432342,
                                slowest=0.420321,
                                elapsed=1,
                                status=ResultStatus.SUCCESS,
                                last_execution_time=datetime.fromtimestamp(1675125600),
                                testcases=[
                                    CaseResult(
                                        name="case01",
                                        status=JudgeStatus.AC,
                                        elapsed=0.2,
                                        memory=30,
                                    ),
                                    CaseResult(
                                        name="case02",
                                        status=JudgeStatus.AC,
                                        elapsed=0.4,
                                        memory=32,
                                    ),
                                ],
                            ),
                            VerificationResult(
                                elapsed=2,
                                status=ResultStatus.SUCCESS,
                                last_execution_time=datetime.fromtimestamp(1675125601),
                            ),
                        ]
                    ),
                    pathlib.Path("hoge.py"): FileResult(
                        verifications=[
                            VerificationResult(
                                heaviest=31.432342,
                                slowest=0.420321,
                                elapsed=1,
                                status=ResultStatus.FAILURE,
                                last_execution_time=datetime.fromtimestamp(1675125600),
                                testcases=[
                                    CaseResult(
                                        name="case01",
                                        status=JudgeStatus.RE,
                                        elapsed=0.2,
                                        memory=30,
                                    ),
                                    CaseResult(
                                        name="case02",
                                        status=JudgeStatus.MLE,
                                        elapsed=0.4,
                                        memory=32,
                                    ),
                                ],
                            ),
                            VerificationResult(
                                elapsed=2,
                                status=ResultStatus.SUCCESS,
                                last_execution_time=datetime.fromtimestamp(1675125601),
                            ),
                        ]
                    ),
                },
            ),
        )

        expected = r"""# ❌ Verification result

- ✔&nbsp;&nbsp;All test case results are `success`
- ❌&nbsp;&nbsp;Test case results containts `failure`
- ⚠&nbsp;&nbsp;Test case results containts `skipped`


## Results
|📝&nbsp;&nbsp;File|✔<br>Passed|❌<br>Failed|⚠<br>Skipped|∑<br>Total|⏳<br>Elapsed|🦥<br>Slowest|🐘<br>Heaviest|
|:---|:---:|:---:|:---:|:---:|:---:|:---:|:---:|
|_**Sum**_|3|3|-|6|4.2s|-|-|
|||||||||
|❌&nbsp;&nbsp;foo/bar.py|-|2|-|2|3.0s|420ms|31.4MB|
|✔&nbsp;&nbsp;foo/baz.py|2|-|-|2|3.0s|420ms|31.4MB|
|❌&nbsp;&nbsp;hoge.py|1|1|-|2|3.0s|420ms|31.4MB|
## Failed tests

### foo/bar.py

|env|name|status|Elapsed|Memory|
|:---|:---|:---:|:---:|:---:|
|ver|case01|TLE|200ms|30MB|
|ver|case02|WA|400ms|32MB|
### hoge.py

|env|name|status|Elapsed|Memory|
|:---|:---|:---:|:---:|:---:|
||case01|RE|200ms|30MB|
||case02|MLE|400ms|32MB|
"""
        assert fp.getvalue() == expected.replace("\r\n", "\n")



from datetime import datetime
from pathlib import Path

import pytest

import competitive_verifier.merge_result.main
from competitive_verifier.models import (
    FileResult,
    ResultStatus,
    VerificationResult,
    VerifyCommandResult,
)

test_merge_params: list[tuple[list[VerifyCommandResult], VerifyCommandResult]] = [
    (
        [
            VerifyCommandResult(
                total_seconds=2.5,
                files={
                    Path("foo"): FileResult(
                        verifications=[
                            VerificationResult(
                                elapsed=1,
                                status=ResultStatus.SUCCESS,
                                last_execution_time=datetime(2020, 2, 26, 23, 45),
                            )
                        ]
                    )
                },
            ),
        ],
        VerifyCommandResult(
            total_seconds=2.5,
            files={
                Path("foo"): FileResult(
                    verifications=[
                        VerificationResult(
                            elapsed=1,
                            status=ResultStatus.SUCCESS,
                            last_execution_time=datetime(2020, 2, 26, 23, 45),
                        )
                    ]
                )
            },
        ),
    ),
    (
        [
            VerifyCommandResult(
                total_seconds=2.5,
                files={
                    Path("foo"): FileResult(
                        verifications=[
                            VerificationResult(
                                elapsed=1,
                                status=ResultStatus.SUCCESS,
                                last_execution_time=datetime(2020, 2, 26, 23, 45),
                            )
                        ]
                    )
                },
            ),
            VerifyCommandResult(
                total_seconds=4.25,
                files={
                    Path("baz"): FileResult(
                        verifications=[
                            VerificationResult(
                                elapsed=1,
                                status=ResultStatus.SKIPPED,
                                last_execution_time=datetime(2021, 2, 27, 23, 45),
                            )
                        ]
                    )
                },
            ),
            VerifyCommandResult(
                total_seconds=1,
                files={
                    Path("bar"): FileResult(
                        verifications=[
                            VerificationResult(
                                elapsed=1,
                                status=ResultStatus.FAILURE,
                                last_execution_time=datetime(2020, 2, 27, 23, 45),
                            )
                        ]
                    )
                },
            ),
        ],
        VerifyCommandResult(
            total_seconds=7.75,
            files={
                Path("foo"): FileResult(
                    verifications=[
                        VerificationResult(
                            elapsed=1,
                            status=ResultStatus.SUCCESS,
                            last_execution_time=datetime(2020, 2, 26, 23, 45),
                        )
                    ]
                ),
                Path("bar"): FileResult(
                    verifications=[
                        VerificationResult(
                            elapsed=1,
                            status=ResultStatus.FAILURE,
                            last_execution_time=datetime(2020, 2, 27, 23, 45),
                        )
                    ]
                ),
                Path("baz"): FileResult(
                    verifications=[
                        VerificationResult(
                            elapsed=1,
                            status=ResultStatus.SKIPPED,
                            last_execution_time=datetime(2021, 2, 27, 23, 45),
                        )
                    ]
                ),
            },
        ),
    ),
]


@pytest.mark.parametrize(
    "results, expected",
    test_merge_params,
    ids=range(len(test_merge_params)),
)
def test_merge(results: list[VerifyCommandResult], expected: VerifyCommandResult):
    assert competitive_verifier.merge_result.main.merge(results) == expected



import pathlib
import re
import shutil
import urllib.parse
from logging import getLogger
from typing import Optional

import yaml

import competitive_verifier.git as git
from competitive_verifier.documents.render import resolve_documentation_of
from competitive_verifier.exec import exec_command
from competitive_verifier.oj.verify.languages.cplusplus import CPlusPlusLanguage
from competitive_verifier.oj.verify.languages.go import GoLanguage
from competitive_verifier.oj.verify.languages.haskell import HaskellLanguage
from competitive_verifier.oj.verify.languages.java import JavaLanguage
from competitive_verifier.oj.verify.languages.nim import NimLanguage
from competitive_verifier.oj.verify.languages.python import PythonLanguage
from competitive_verifier.oj.verify.languages.ruby import RubyLanguage
from competitive_verifier.oj.verify.languages.rust import RustLanguage
from competitive_verifier.oj.verify.list import OjVerifyConfig
from competitive_verifier.oj.verify.models import Language

logger = getLogger(__name__)


_title_pattern = re.compile(r"@(?:title|brief) (.*)")
_docs_pattern = re.compile(r"@docs (.*)$", re.MULTILINE)
_documentation_of_pattern = re.compile(r"^documentation_of:.*", re.MULTILINE)

_problem_pattern = re.compile(r"#define PROBLEM(?:\\\n| |\t)+(.+)")
_error_pattern = re.compile(r"#define ERROR(?:\\\n| |\t)+(.+)")


def problem_subn(content: str) -> tuple[str, int]:
    return _problem_pattern.subn(
        lambda m: f"// competitive-verifier: PROBLEM {_strip_quote(m.group(1))}",
        content,
    )


def error_subn(content: str) -> tuple[str, int]:
    return _error_pattern.subn(
        lambda m: f"// competitive-verifier: ERROR {_strip_quote(m.group(1))}", content
    )


_yukicoder_pattern = re.compile(r"[^#]*YUKICODER_TOKEN: .*")


def migrate_conf_dir(*, dry_run: bool):
    old_conf_dir = pathlib.Path(".verify-helper")
    new_conf_dir = pathlib.Path(".competitive-verifier")
    old_docs_dir = old_conf_dir / "docs"
    new_docs_dir = new_conf_dir / "docs"
    if not old_docs_dir.exists():
        return

    if not new_docs_dir.exists():
        gitignore = new_conf_dir / ".gitignore"
        if not dry_run:
            new_docs_dir.mkdir(parents=True, exist_ok=True)
            gitignore.write_bytes(b"/*/\n!docs/")
        logger.warning("Create directory: %s", new_docs_dir.as_posix())

    for p in old_docs_dir.glob("**/*"):
        relative = p.relative_to(old_docs_dir)
        new_path = new_docs_dir / relative
        if p.is_dir():
            logger.warning("Create directory: %s", new_path.as_posix())
            new_path.mkdir(parents=True, exist_ok=True)
        elif p.is_file():
            logger.warning("Move docs file: %s → %s", p.as_posix(), new_path.as_posix())
            if not dry_run:
                new_path.parent.mkdir(parents=True, exist_ok=True)
                shutil.copyfile(p, new_path)
                p.unlink(missing_ok=True)
    shutil.rmtree(old_docs_dir)


def _get_docs_path(content: str, *, path: pathlib.Path) -> Optional[pathlib.Path]:
    docs_match = _docs_pattern.search(content)
    if docs_match:
        doc_path = docs_match.group(1)
        if isinstance(doc_path, str):
            return resolve_documentation_of(
                doc_path.strip(),
                basedir=path.parent,
            )
    return None


def _strip_quote(s: str) -> str:
    return s.strip().strip("\"'")


def migrate_cpp_annotations(path: pathlib.Path, *, dry_run: bool):
    hit = False
    logger.debug("Migrate file: %s", path.as_posix())
    content = path.read_text(encoding="utf-8")

    new_content, hit_cnt = problem_subn(content)
    if hit_cnt > 0:
        logger.warning(
            "[Updated] %s: Replace `#define PROBLEM` to `competitive-verifier: PROBLEM`",
            path.as_posix(),
        )
        content = new_content
        hit = True

    new_content, hit_cnt = error_subn(content)
    if hit_cnt > 0:
        logger.warning(
            "[Updated] %s: Replace `#define ERROR` to `competitive-verifier: ERROR`",
            path.as_posix(),
        )
        content = new_content
        hit = True

    doc_path = _get_docs_path(content, path=path)
    if doc_path:
        if not doc_path.exists():
            logger.error("`%s` doesn't exits.", doc_path.as_posix())

        docs = doc_path.read_text("utf-8")
        docs_lines = docs.splitlines(keepends=True)

        logger.warning("[Updated] %s: Remove `@docs`", path.as_posix())
        content, _ = _docs_pattern.subn("", content)
        hit = True

        if _documentation_of_pattern.search(docs):
            logger.debug(
                "[Not Updated] %s: Add `documentation_of:`",
                doc_path.as_posix(),
            )
        else:
            logger.warning(
                "[Updated] %s: Add `documentation_of:`",
                doc_path.as_posix(),
            )
            documentation_of = f"documentation_of: //{path.as_posix()}\n"
            if docs.startswith("---"):
                docs_lines.insert(1, documentation_of)
            else:
                docs_lines[0:0] = [
                    "---\n",
                    documentation_of,
                    "---\n\n",
                ]
            if not dry_run:
                doc_path.write_text("".join(docs_lines), encoding="utf-8")

    if hit:
        if not dry_run:
            path.write_text(content, "utf-8")
    else:
        logger.debug("Not updated: %s", path.as_posix())


def _lang_type_to_str(lang: Optional[Language]) -> Optional[str]:
    if isinstance(lang, CPlusPlusLanguage):
        return "cpp"
    if isinstance(lang, GoLanguage):
        return "go"
    if isinstance(lang, HaskellLanguage):
        return "haskel"
    if isinstance(lang, JavaLanguage):
        return "java"
    if isinstance(lang, NimLanguage):
        return "nim"
    if isinstance(lang, PythonLanguage):
        return "python"
    if isinstance(lang, RubyLanguage):
        return "ruby"
    if isinstance(lang, RustLanguage):
        return "rust"
    return None


def _get_action_query(languages: set[str]) -> dict[str, str]:
    d = {
        "configToml": ".verify-helper/config.toml",
        "langs": "|".join(languages),
    }
    if not pathlib.Path(d["configToml"]).exists():
        del d["configToml"]
    remote = exec_command(
        ["git", "remote", "get-url", "origin"],
        text=True,
        capture_output=True,
    ).stdout.strip()
    if remote:
        d["repository"] = remote

    workflow_path = pathlib.Path(".github/workflows/verify.yml")
    if workflow_path.exists():
        workflow = workflow_path.read_text()
        if _yukicoder_pattern.search(workflow):
            d["tokens"] = "yuki"

    jekyll_config_path = pathlib.Path(".competitive-verifier/docs/_config.yml")
    if jekyll_config_path.exists():
        with jekyll_config_path.open("r") as fp:
            jekyll_config = yaml.safe_load(fp)
        try:
            exclude = jekyll_config.get("exclude")
            if isinstance(exclude, list):
                exclude = "\n".join(
                    exclude  # pyright: ignore[reportUnknownArgumentType]
                )
            if exclude:
                d["exclude"] = exclude
        except Exception:
            pass
    return d


def main(dry_run: bool) -> bool:
    migrate_conf_dir(dry_run=dry_run)

    languages = set[str]()
    lang_dict = OjVerifyConfig().get_dict()

    for path in git.ls_files():
        lang = lang_dict.get(path.suffix)
        if isinstance(lang, CPlusPlusLanguage):
            migrate_cpp_annotations(path, dry_run=dry_run)
        lang_str = _lang_type_to_str(lang)
        if lang_str:
            languages.add(lang_str)

    d = _get_action_query(languages)

    logger.info("Complete migrations")
    print("Next steps")
    # page = "http://localhost:4000"
    page = "https://competitive-verifier.github.io/competitive-verifier"
    print(f"  1. Open {page}/installer.html?{urllib.parse.urlencode(d)}")
    print("  2. Update your GitHub Actions.")
    return True



import pathlib

import pytest
from pytest_mock import MockerFixture


def pytest_addoption(parser: pytest.Parser):
    parser.addoption(
        "--use-prev-dest",
        action="store_true",
        help="Skip deletion of dst_dir.",
    )


@pytest.fixture
def mock_exists(mocker: MockerFixture):
    def _mock_exists(val: bool):
        return mocker.patch.object(pathlib.Path, "exists", return_value=val)

    return _mock_exists


@pytest.fixture
def mock_exec_command(mocker: MockerFixture):
    return mocker.patch("subprocess.run")



import datetime
import json
import pathlib
import random
import shutil
from typing import Any, Iterable, List

import onlinejudge.dispatch
import onlinejudge.service.library_checker as library_checker
import pytest
import requests
from onlinejudge.type import TestCase as OjTestCase
from pytest_mock import MockerFixture

import competitive_verifier.config as config
from competitive_verifier.models import FileResult, JudgeStatus, ResultStatus
from competitive_verifier.models import TestcaseResult as _TestcaseResult
from competitive_verifier.models import VerificationResult
from competitive_verifier.verify import main, verifier
from tests.integration.utils import md5_number

from .data.integration_data import IntegrationData


@pytest.mark.usefixtures("additional_path")
@pytest.mark.order(-500)
class TestCommandVerfy:
    @pytest.mark.usefixtures("mock_verification")
    def test_mock_dump(self):
        for _ in range(20):
            command_result = verifier.VerifyCommandResult(
                total_seconds=3,
                files={
                    pathlib.Path("foo/bar.py"): FileResult(
                        verifications=[
                            VerificationResult(
                                verification_name="name1",
                                elapsed=random.randint(10, 242),
                                status=ResultStatus.SUCCESS,
                                last_execution_time=datetime.datetime.fromtimestamp(
                                    random.randint(1000000000, 3000000000)
                                ),
                            ),
                            VerificationResult(
                                verification_name="name2",
                                elapsed=random.randint(10, 242),
                                status=ResultStatus.SKIPPED,
                                heaviest=random.randint(1000, 3000),
                                slowest=random.randint(234, 1234),
                                last_execution_time=datetime.datetime.fromtimestamp(
                                    random.randint(1000000000, 3000000000)
                                ),
                                testcases=[
                                    _TestcaseResult(
                                        name="case01",
                                        status=JudgeStatus.AC,
                                        elapsed=random.randint(12, 24),
                                        memory=random.randint(100, 1000),
                                    ),
                                    _TestcaseResult(
                                        name="case02",
                                        status=JudgeStatus.RE,
                                        elapsed=random.randint(12, 24),
                                        memory=random.randint(100, 1000),
                                    ),
                                    _TestcaseResult(
                                        name="case03",
                                        status=JudgeStatus.WA,
                                        elapsed=random.randint(12, 24),
                                        memory=random.randint(100, 1000),
                                    ),
                                    _TestcaseResult(
                                        name="case04",
                                        status=JudgeStatus.TLE,
                                        elapsed=random.randint(12, 24),
                                        memory=random.randint(100, 1000),
                                    ),
                                ],
                            ),
                        ]
                    ),
                    pathlib.Path("foo/baz2.py"): FileResult(
                        verifications=[
                            VerificationResult(
                                verification_name="name1",
                                elapsed=random.randint(10, 242),
                                status=ResultStatus.SUCCESS,
                                last_execution_time=datetime.datetime.fromtimestamp(
                                    random.randint(1000000000, 3000000000)
                                ),
                            ),
                        ]
                    ),
                },
            )
            assert json.loads(command_result.model_dump_json(exclude_none=True)) == {
                "total_seconds": 2547.12,
                "files": {
                    "foo/bar.py": {
                        "verifications": [
                            {
                                "verification_name": "name1",
                                "status": "success",
                                "elapsed": 9948.0,
                                "last_execution_time": "2004-10-09T00:26:13.890000+02:00",
                            },
                            {
                                "verification_name": "name2",
                                "status": "skipped",
                                "elapsed": 2238.0,
                                "slowest": 223.0,
                                "heaviest": 753.0,
                                "testcases": [
                                    {
                                        "name": "case01",
                                        "status": "AC",
                                        "elapsed": 1.33,
                                        "memory": 32.38,
                                    },
                                    {
                                        "name": "case02",
                                        "status": "RE",
                                        "elapsed": 4.73,
                                        "memory": 9.88,
                                    },
                                    {
                                        "name": "case03",
                                        "status": "WA",
                                        "elapsed": 2.4,
                                        "memory": 35.43,
                                    },
                                    {
                                        "name": "case04",
                                        "status": "TLE",
                                        "elapsed": 2.42,
                                        "memory": 22.41,
                                    },
                                ],
                                "last_execution_time": "2026-11-12T22:34:03.760000-11:00",
                            },
                        ],
                        "newest": True,
                    },
                    "foo/baz2.py": {
                        "verifications": [
                            {
                                "verification_name": "name1",
                                "status": "success",
                                "elapsed": 1239.0,
                                "last_execution_time": "1975-10-05T09:39:58.780000-09:00",
                            }
                        ],
                        "newest": True,
                    },
                },
            }

    @pytest.mark.each_language_integration
    @pytest.mark.integration
    @pytest.mark.usefixtures("mock_verification")
    def test_verify(
        self,
        integration_data: IntegrationData,
    ):
        verify = integration_data.config_dir_path / "verify.json"
        result = integration_data.config_dir_path / "result.json"
        main.main(["--verify-json", str(verify), "--output", str(result)])

        assert (
            json.loads(pathlib.Path(result).read_bytes())
            == integration_data.expected_verify_result()
        )


class _MockVerifyCommandResult(verifier.VerifyCommandResult):
    def model_dump_json(
        self,
        *,
        indent: Any = None,
        include: Any = None,
        exclude: Any = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        round_trip: bool = False,
        warnings: bool = True,
    ) -> str:
        return self.model_copy()._dump_super(
            indent=indent,
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            round_trip=round_trip,
            warnings=warnings,
        )

    def _dump_super(
        self,
        *,
        indent: Any = None,
        include: Any = None,
        exclude: Any = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        round_trip: bool = False,
        warnings: bool = True,
    ) -> str:
        def rewriteVerifyCommandResult(result: verifier.VerifyCommandResult):
            result.total_seconds = len(result.files) * 1234.56 + 78
            result.files = {k: rewriteFileResult(k, v) for k, v in result.files.items()}
            return result

        def rewriteFileResult(path: pathlib.Path, file_result: FileResult):
            seed = path.as_posix().encode()
            file_result.verifications = [
                rewriteVerificationResult(seed, v) for v in file_result.verifications
            ]
            return file_result

        def rewriteVerificationResult(seed: bytes, verification: VerificationResult):
            seed += (verification.verification_name or "").encode()
            seed += verification.status.name.encode()

            verification.elapsed = md5_number(seed + b"elapsed") % 10000

            if verification.slowest:
                verification.slowest = verification.elapsed // 10
            if verification.heaviest:
                verification.heaviest = md5_number(seed + b"heaviest") % 1000

            if verification.testcases:
                verification.testcases = [
                    rewriteTestcaseResult(seed, c) for c in verification.testcases
                ]

            random_time = md5_number(seed + b"last_execution_time")
            verification.last_execution_time = datetime.datetime.fromtimestamp(
                random_time % 300000000000 / 100,
                tz=datetime.timezone(datetime.timedelta(hours=random_time % 25 - 12)),
            )

            return verification

        def rewriteTestcaseResult(seed: bytes, case: _TestcaseResult):
            seed += (case.name or "").encode()
            seed += case.status.name.encode()

            case.elapsed = md5_number(seed + b"elapsed") % 1000 / 100
            case.memory = md5_number(seed + b"memory") % 10000 / 100
            return case

        rewriteVerifyCommandResult(self)
        return super().model_dump_json(
            indent=indent,
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            round_trip=round_trip,
            warnings=warnings,
        )


class _MockLibraryCheckerProblem(library_checker.LibraryCheckerProblem):
    def __init__(self, *, problem_id: str):
        super().__init__(problem_id=problem_id)

    def download_system_cases(self, *, session: Any = None) -> List[OjTestCase]:
        self._update_cloned_repository()
        return list(self._mock_cases())

    def download_sample_cases(self, *, session: Any = None) -> List[OjTestCase]:
        assert False

    def _update_cloned_repository(self) -> None:
        zip_path = config.get_cache_dir() / "library-checker-problems.zip"
        if not zip_path.exists():
            zip_path.parent.mkdir(parents=True, exist_ok=True)
            res = requests.get(
                "https://github.com/yosupo06/library-checker-problems/archive/refs/heads/master.zip"
            )
            with zip_path.open("wb") as fp:
                fp.write(res.content)
            shutil.unpack_archive(zip_path, config.get_cache_dir())
            shutil.move(
                config.get_cache_dir() / "library-checker-problems-master",
                config.get_cache_dir() / "library-checker-problems",
            )

        # library_checker.LibraryCheckerService._update_cloned_repository()  # pyright: ignore[reportPrivateUsage]

    def _mock_cases(self) -> Iterable[OjTestCase]:
        if self.problem_id == "aplusb":
            for name, a, b in [
                ("example_00", 1000, 10),
                ("example_01", 1002, 20),
                ("random_00", 1000, 130),
                ("random_01", 1003, 140),
                ("random_02", 1005, 150),
                ("random_03", 1000, 160),
                ("random_04", 1005, 170),
                ("random_05", 1007, 180),
                ("random_06", 1009, 191),
                ("random_07", 1008, 200),
                ("random_08", 1005, 214),
                ("random_09", 1008, 225),
            ]:
                yield OjTestCase(
                    name=name,
                    input_name=f"{name}.in",
                    output_name=f"{name}.out",
                    input_data=f"{a} {b}\n".encode(),
                    output_data=f"{a+b}\n".encode(),
                )
        else:
            raise NotImplementedError()


@pytest.fixture
def mock_verification(mocker: MockerFixture):
    mocker.patch(
        "competitive_verifier.verify.verifier.VerifyCommandResult",
        side_effect=_MockVerifyCommandResult,
    )

    mocker.patch.object(
        onlinejudge.dispatch,
        "problems",
        new=[
            _MockLibraryCheckerProblem
            if p == library_checker.LibraryCheckerProblem
            else p
            for p in onlinejudge.dispatch.problems
        ],
    )



# pyright: reportPrivateUsage=none
import datetime
from pathlib import Path
from typing import Any, Callable, Optional

import pytest

from competitive_verifier.models import (
    CommandVerification,
    ConstVerification,
    FileResult,
    ResultStatus,
    VerificationFile,
    VerificationInput,
    VerificationResult,
    VerifyCommandResult,
)
from competitive_verifier.verify.verifier import InputContainer, SplitState

SUCCESS = ResultStatus.SUCCESS


class MockInputContainer(InputContainer):
    def __init__(
        self,
        obj: Any = None,
        *,
        prev_result: Optional[VerifyCommandResult] = None,
        verification_time: Optional[datetime.datetime] = None,
        file_timestamps: dict[Optional[Path], datetime.datetime] = {},
        split_state: Optional[SplitState] = None,
    ) -> None:
        super().__init__(
            input=VerificationInput.model_validate(obj) if obj else VerificationInput(),
            verification_time=verification_time or datetime.datetime.now(),
            prev_result=prev_result,
            split_state=split_state,
        )

        self.file_timestamps = file_timestamps

    def get_file_timestamp(self, path: Path) -> datetime.datetime:
        assert self.file_timestamps is not None
        dt = self.file_timestamps.get(path)
        if dt:
            return dt
        return self.file_timestamps[None]


test_verification_files_params: list[
    tuple[InputContainer, dict[Path, VerificationFile]]
] = [
    (
        MockInputContainer(
            {
                "files": {
                    "foo": {},
                    "bar": {},
                    "baz": {},
                },
            }
        ),
        {},
    ),
    (
        MockInputContainer(
            {
                "files": {
                    "foo": {
                        "verification": {
                            "type": "const",
                            "status": "success",
                        }
                    },
                    "bar": {
                        "verification": {
                            "type": "const",
                            "status": "success",
                        }
                    },
                    "baz": {
                        "verification": {
                            "type": "const",
                            "status": "success",
                        }
                    },
                },
            }
        ),
        {
            Path("foo"): VerificationFile(
                verification=ConstVerification(status=SUCCESS),
            ),
            Path("bar"): VerificationFile(
                verification=ConstVerification(status=SUCCESS),
            ),
            Path("baz"): VerificationFile(
                verification=ConstVerification(status=SUCCESS),
            ),
        },
    ),
    (
        MockInputContainer(
            {
                "files": {
                    "foo": {
                        "verification": {
                            "type": "const",
                            "status": "success",
                        }
                    },
                    "bar": {},
                    "baz": {
                        "verification": {
                            "type": "const",
                            "status": "success",
                        }
                    },
                },
            }
        ),
        {
            Path("foo"): VerificationFile(
                verification=ConstVerification(status=SUCCESS),
            ),
            Path("baz"): VerificationFile(
                verification=ConstVerification(status=SUCCESS),
            ),
        },
    ),
]


@pytest.mark.parametrize(
    "resolver, expected",
    test_verification_files_params,
    ids=range(len(test_verification_files_params)),
)
def test_verification_files(
    resolver: InputContainer,
    expected: dict[Path, VerificationFile],
):
    assert resolver.verification_files == expected


test_file_need_verification_params: list[
    tuple[InputContainer, Path, FileResult, bool]
] = [
    (
        MockInputContainer(
            verification_time=datetime.datetime(2018, 12, 25),
            file_timestamps={
                Path("foo"): datetime.datetime(2015, 12, 25),
            },
        ),
        Path("foo"),
        FileResult(
            verifications=[
                VerificationResult(
                    elapsed=1.5,
                    status=ResultStatus.SUCCESS,
                    last_execution_time=datetime.datetime(2016, 12, 24),
                ),
            ]
        ),
        False,
    ),
    (
        MockInputContainer(
            verification_time=datetime.datetime(2018, 12, 25),
            file_timestamps={
                Path("foo"): datetime.datetime(2017, 12, 25),
            },
        ),
        Path("foo"),
        FileResult(
            verifications=[
                VerificationResult(
                    elapsed=1.5,
                    status=ResultStatus.SUCCESS,
                    last_execution_time=datetime.datetime(2016, 12, 24),
                ),
            ]
        ),
        True,
    ),
    (
        MockInputContainer(
            verification_time=datetime.datetime(2018, 12, 25),
            file_timestamps={
                Path("foo"): datetime.datetime(2015, 12, 25),
            },
        ),
        Path("foo"),
        FileResult(
            verifications=[
                VerificationResult(
                    elapsed=1.5,
                    status=ResultStatus.FAILURE,
                    last_execution_time=datetime.datetime(2016, 12, 24),
                ),
            ]
        ),
        True,
    ),
    (
        MockInputContainer(
            verification_time=datetime.datetime(2018, 12, 25),
            file_timestamps={
                Path("foo"): datetime.datetime(2015, 12, 25),
            },
        ),
        Path("foo"),
        FileResult(
            verifications=[
                VerificationResult(
                    elapsed=1.5,
                    status=ResultStatus.SKIPPED,
                    last_execution_time=datetime.datetime(2016, 12, 24),
                ),
            ]
        ),
        True,
    ),
    (
        MockInputContainer(
            verification_time=datetime.datetime(2018, 12, 25),
            file_timestamps={
                Path("foo"): datetime.datetime(2015, 12, 25),
            },
        ),
        Path("foo"),
        FileResult(
            verifications=[
                VerificationResult(
                    elapsed=1.5,
                    status=ResultStatus.SUCCESS,
                    last_execution_time=datetime.datetime(2016, 12, 24),
                ),
            ]
        ),
        False,
    ),
    (
        MockInputContainer(
            verification_time=datetime.datetime(2015, 12, 25),
            file_timestamps={
                Path("foo"): datetime.datetime(2017, 12, 25),
            },
        ),
        Path("foo"),
        FileResult(
            verifications=[
                VerificationResult(
                    elapsed=1.5,
                    status=ResultStatus.SUCCESS,
                    last_execution_time=datetime.datetime(2016, 12, 24),
                ),
            ]
        ),
        False,
    ),
]


@pytest.mark.parametrize(
    "resolver, path, file_result, expected",
    test_file_need_verification_params,
    ids=range(len(test_file_need_verification_params)),
)
def test_file_need_verification(
    resolver: InputContainer,
    path: Path,
    file_result: FileResult,
    expected: bool,
    mock_exists: Callable[[bool], Any],
):
    mock_exists(True)
    assert resolver.file_need_verification(path, file_result) == expected


@pytest.mark.parametrize(
    "resolver, path, file_result, _",
    test_file_need_verification_params,
    ids=range(len(test_file_need_verification_params)),
)
def test_file_need_verification_no_file(
    resolver: InputContainer,
    path: Path,
    file_result: FileResult,
    _: bool,
    mock_exists: Callable[[bool], Any],
):
    mock_exists(False)
    assert not resolver.file_need_verification(path, file_result)


test_remaining_verification_files_params: list[
    tuple[InputContainer, dict[Path, VerificationFile]]
] = [
    (
        MockInputContainer(
            {
                "files": {
                    "foo": {
                        "verification": {
                            "type": "const",
                            "status": "success",
                        }
                    },
                    "bar": {
                        "verification": {
                            "type": "const",
                            "status": "success",
                        }
                    },
                    "baz": {
                        "verification": {
                            "type": "const",
                            "status": "success",
                        }
                    },
                },
            }
        ),
        {},
    ),
    (
        MockInputContainer(
            {
                "files": {
                    "foo": {"verification": {"type": "command", "command": "true"}},
                    "bar": {
                        "verification": {
                            "type": "const",
                            "status": "success",
                        }
                    },
                    "baz": {"verification": {"type": "command", "command": "true"}},
                },
            }
        ),
        {
            Path("foo"): VerificationFile(
                verification=CommandVerification(command="true"),
            ),
            Path("baz"): VerificationFile(
                verification=CommandVerification(command="true"),
            ),
        },
    ),
    (
        MockInputContainer(
            {
                "files": {
                    "foo": {"verification": {"type": "command", "command": "true"}},
                    "bar": {
                        "verification": {
                            "type": "const",
                            "status": "success",
                        }
                    },
                    "baz": {"verification": {"type": "command", "command": "true"}},
                },
            },
            verification_time=datetime.datetime(9999, 5, 22),
            file_timestamps={
                None: datetime.datetime(2018, 5, 22),
            },
            prev_result=VerifyCommandResult(
                total_seconds=1.5,
                files={
                    Path("baz"): FileResult(
                        verifications=[
                            VerificationResult(
                                elapsed=1.5,
                                status=ResultStatus.SUCCESS,
                                last_execution_time=datetime.datetime(2018, 7, 22),
                            )
                        ]
                    )
                },
            ),
        ),
        {
            Path("foo"): VerificationFile(
                verification=CommandVerification(command="true"),
            ),
        },
    ),
]


@pytest.mark.parametrize(
    "resolver, expected",
    test_remaining_verification_files_params,
    ids=range(len(test_remaining_verification_files_params)),
)
def test_remaining_verification_files(
    resolver: InputContainer,
    expected: dict[Path, VerificationFile],
):
    assert resolver.remaining_verification_files == expected


test_current_verification_files_params: list[
    tuple[int, dict[Path, VerificationFile]]
] = [
    (
        0,
        {
            Path("bar/0.py"): VerificationFile(
                verification=CommandVerification(command="true"),
            ),
            Path("bar/1.py"): VerificationFile(
                verification=CommandVerification(command="true"),
            ),
            Path("bar/2.py"): VerificationFile(
                verification=CommandVerification(command="true"),
            ),
        },
    ),
    (
        1,
        {
            Path("bar/3.py"): VerificationFile(
                verification=CommandVerification(command="true"),
            ),
            Path("baz/0.py"): VerificationFile(
                verification=CommandVerification(command="true"),
            ),
            Path("baz/1.py"): VerificationFile(
                verification=CommandVerification(command="true"),
            ),
        },
    ),
    (
        2,
        {
            Path("baz/2.py"): VerificationFile(
                verification=CommandVerification(command="true"),
            ),
            Path("baz/3.py"): VerificationFile(
                verification=CommandVerification(command="true"),
            ),
            Path("foo/0.py"): VerificationFile(
                verification=CommandVerification(command="true"),
            ),
        },
    ),
    (
        3,
        {
            Path("foo/1.py"): VerificationFile(
                verification=CommandVerification(command="true"),
            ),
            Path("foo/2.py"): VerificationFile(
                verification=CommandVerification(command="true"),
            ),
            Path("foo/3.py"): VerificationFile(
                verification=CommandVerification(command="true"),
            ),
        },
    ),
    (
        4,
        {},
    ),
]


@pytest.mark.parametrize(
    "index, expected",
    test_current_verification_files_params,
    ids=range(len(test_current_verification_files_params)),
)
def test_current_verification_files(index: int, expected: dict[Path, VerificationFile]):
    command_verification = {"verification": {"type": "command", "command": "true"}}
    resolver = MockInputContainer(
        {
            "files": {
                "dummy/0.py": {},
                "foo/0.py": command_verification,
                "bar/0.py": command_verification,
                "baz/0.py": command_verification,
                "dummy/1.py": {},
                "foo/1.py": command_verification,
                "bar/1.py": command_verification,
                "baz/1.py": command_verification,
                "foo/2.py": command_verification,
                "dummy/2.py": {},
                "bar/2.py": command_verification,
                "baz/2.py": command_verification,
                "baz/3.py": command_verification,
                "dummy/3.py": {},
                "bar/3.py": command_verification,
                "foo/3.py": command_verification,
            },
        },
        split_state=SplitState(size=4, index=index),
    )
    remaining_verification_files = {
        Path("foo/0.py"): VerificationFile(
            verification=CommandVerification(command="true"),
        ),
        Path("bar/0.py"): VerificationFile(
            verification=CommandVerification(command="true"),
        ),
        Path("baz/0.py"): VerificationFile(
            verification=CommandVerification(command="true"),
        ),
        Path("foo/1.py"): VerificationFile(
            verification=CommandVerification(command="true"),
        ),
        Path("bar/1.py"): VerificationFile(
            verification=CommandVerification(command="true"),
        ),
        Path("baz/1.py"): VerificationFile(
            verification=CommandVerification(command="true"),
        ),
        Path("foo/2.py"): VerificationFile(
            verification=CommandVerification(command="true"),
        ),
        Path("bar/2.py"): VerificationFile(
            verification=CommandVerification(command="true"),
        ),
        Path("baz/2.py"): VerificationFile(
            verification=CommandVerification(command="true"),
        ),
        Path("foo/3.py"): VerificationFile(
            verification=CommandVerification(command="true"),
        ),
        Path("bar/3.py"): VerificationFile(
            verification=CommandVerification(command="true"),
        ),
        Path("baz/3.py"): VerificationFile(
            verification=CommandVerification(command="true"),
        ),
    }
    assert resolver.remaining_verification_files == remaining_verification_files
    assert resolver.current_verification_files == expected



import pathlib
from subprocess import CompletedProcess
from typing import Annotated, Literal, Optional, Union, overload

from pydantic import BaseModel, Field

from competitive_verifier.exec import exec_command


class ShellCommand(BaseModel):
    command: Union[list[str], str] = Field(
        description="Shell command",
    )
    """Shell command
    """

    env: Optional[dict[str, str]] = Field(
        default=None,
        description="Envitonment variables for command",
    )
    """Envitonment variables for command
    """

    cwd: Optional[pathlib.Path] = Field(
        default=None,
        description="The working directory of child process.",
    )
    """The working directory of child process.
    """

    @overload
    def exec_command(
        self,
        text: Literal[False] = False,
        check: bool = False,
        capture_output: bool = False,
        group_log: bool = False,
    ) -> CompletedProcess[bytes]:
        ...

    @overload
    def exec_command(
        self,
        text: Literal[True] = True,
        check: bool = False,
        capture_output: bool = False,
        group_log: bool = False,
    ) -> CompletedProcess[str]:
        ...

    def exec_command(
        self,
        text: bool = False,
        check: bool = False,
        capture_output: bool = False,
        group_log: bool = False,
    ) -> Union[CompletedProcess[str], CompletedProcess[bytes]]:
        return exec_command(
            command=self.command,
            env=self.env,
            text=text,
            check=check,
            capture_output=capture_output,
            cwd=self.cwd,
            group_log=group_log,
        )

    @classmethod
    def parse_command_like(cls, cmd: "ShellCommandLike") -> "ShellCommand":
        if isinstance(cmd, str) or isinstance(cmd, list):
            return ShellCommand(command=cmd)
        return cmd


ShellCommandLike = Annotated[
    Union[ShellCommand, Union[list[str], str]],
    Field(
        examples=[
            "command",
            ["command", "arg1", "arg2"],
            ShellCommand(
                command=["command", "arg1", "arg2"],
                env={"ENVVAR": "DUMMY"},
                cwd=pathlib.Path("/tmp"),
            ),
            ShellCommand(
                command="command",
                env={"ENVVAR": "DUMMY"},
                cwd=pathlib.Path("/tmp"),
            ),
        ]
    ),
]



import hashlib
import pathlib
import sys
from typing import Optional

from onlinejudge.service.atcoder import AtCoderService
from onlinejudge.service.library_checker import LibraryCheckerProblem
from onlinejudge.service.yukicoder import YukicoderService

import competitive_verifier.config as config

checker_exe_name = "checker.exe" if sys.platform == "win32" else "checker"


def get_cache_directory() -> pathlib.Path:
    return config.get_cache_dir().resolve() / "online-judge-tools"


def get_problem_cache_dir() -> pathlib.Path:
    return config.get_cache_dir() / "problems"


def get_directory(url: str) -> pathlib.Path:
    return get_problem_cache_dir() / hashlib.md5(url.encode()).hexdigest()


def is_yukicoder(url: str) -> bool:
    return YukicoderService.from_url(url) is not None


def is_atcoder(url: str) -> bool:
    return AtCoderService.from_url(url) is not None


def get_checker_problem(url: str) -> Optional[LibraryCheckerProblem]:
    return LibraryCheckerProblem.from_url(url)


def get_checker_path(url: str) -> Optional[pathlib.Path]:
    checker_problem = get_checker_problem(url)
    if checker_problem:
        problem_dir = (
            checker_problem._get_problem_directory_path()  # pyright: ignore[reportPrivateUsage]
        )
        return problem_dir / checker_exe_name
